@@ -8,7 +8,13 @@
 # -----------------------------------------------------------------------------
 # Imports
 # -----------------------------------------------------------------------------
+from __future__ import annotations
+
 import abc
+from typing import TYPE_CHECKING, Any
+
+if TYPE_CHECKING:
+    from .channelsabc import ChannelABC
 
 # -----------------------------------------------------------------------------
 # Main kernel client class
@@ -24,27 +30,27 @@ class KernelClientABC(metaclass=abc.ABCMeta):
     """
 
     @abc.abstractproperty
-    def kernel(self):
+    def kernel(self) -> Any:
         pass
 
     @abc.abstractproperty
-    def shell_channel_class(self):
+    def shell_channel_class(self) -> type[ChannelABC]:
         pass
 
     @abc.abstractproperty
-    def iopub_channel_class(self):
+    def iopub_channel_class(self) -> type[ChannelABC]:
         pass
 
     @abc.abstractproperty
-    def hb_channel_class(self):
+    def hb_channel_class(self) -> type[ChannelABC]:
         pass
 
     @abc.abstractproperty
-    def stdin_channel_class(self):
+    def stdin_channel_class(self) -> type[ChannelABC]:
         pass
 
     @abc.abstractproperty
-    def control_channel_class(self):
+    def control_channel_class(self) -> type[ChannelABC]:
         pass
 
     # --------------------------------------------------------------------------
@@ -52,36 +58,43 @@ class KernelClientABC(metaclass=abc.ABCMeta):
     # --------------------------------------------------------------------------
 
     @abc.abstractmethod
-    def start_channels(self, shell=True, iopub=True, stdin=True, hb=True, control=True):
+    def start_channels(
+        self,
+        shell: bool = True,
+        iopub: bool = True,
+        stdin: bool = True,
+        hb: bool = True,
+        control: bool = True,
+    ) -> None:
         """Start the channels for the client."""
         pass
 
     @abc.abstractmethod
-    def stop_channels(self):
+    def stop_channels(self) -> None:
         """Stop the channels for the client."""
         pass
 
     @abc.abstractproperty
-    def channels_running(self):
+    def channels_running(self) -> bool:
         """Get whether the channels are running."""
         pass
 
     @abc.abstractproperty
-    def shell_channel(self):
+    def shell_channel(self) -> ChannelABC:
         pass
 
     @abc.abstractproperty
-    def iopub_channel(self):
+    def iopub_channel(self) -> ChannelABC:
         pass
 
     @abc.abstractproperty
-    def stdin_channel(self):
+    def stdin_channel(self) -> ChannelABC:
         pass
 
     @abc.abstractproperty
-    def hb_channel(self):
+    def hb_channel(self) -> ChannelABC:
         pass
 
     @abc.abstractproperty
-    def control_channel(self):
+    def control_channel(self) -> ChannelABC:
         pass
