@@ -5,6 +5,8 @@ related to writing and reading connections files.
 """
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+from __future__ import annotations
+
 import errno
 import glob
 import json
@@ -15,7 +17,7 @@ import tempfile
 import warnings
 import sys
 from getpass import getpass
-from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast
+from typing import TYPE_CHECKING, Any, Dict, Union, cast
 
 import zmq
 from jupyter_core.paths import jupyter_data_dir, jupyter_runtime_dir, secure_write
@@ -25,12 +27,17 @@ from traitlets.config import LoggingConfigurable, SingletonConfigurable
 from .localinterfaces import localhost
 from .utils import _filefind
 
+if TYPE_CHECKING:
+    from jupyter_client import BlockingKernelClient
+
+    from .session import Session
+
 # Define custom type for kernel connection info
 KernelConnectionInfo = Dict[str, Union[int, str, bytes]]
 
 
 def write_connection_file(
-    fname: Optional[str] = None,
+    fname: str | None = None,
     shell_port: int = 0,
     iopub_port: int = 0,
     stdin_port: int = 0,
@@ -41,7 +48,8 @@ def write_connection_file(
     transport: str = "tcp",
     signature_scheme: str = "hmac-sha256",
     kernel_name: str = "",
-) -> Tuple[str, KernelConnectionInfo]:
+    **kwargs: Any,
+) -> tuple[str, KernelConnectionInfo]:
     """Generates a JSON config file, including the selection of random ports.
 
     Parameters
@@ -91,8 +99,8 @@ def write_connection_file(
 
     # Find open ports as necessary.
 
-    ports: List[int] = []
-    sockets: List[socket.socket] = []
+    ports: list[int] = []
+    sockets: list[socket.socket] = []
     ports_needed = (
         int(shell_port <= 0)
         + int(iopub_port <= 0)
@@ -114,7 +122,7 @@ def write_connection_file(
     else:
         N = 1
         for _ in range(ports_needed):
-            while os.path.exists(f"{ip}-{str(N)}"):
+            while os.path.exists(f"{ip}-{N!s}"):
                 N += 1
             ports.append(N)
             N += 1
@@ -141,6 +149,7 @@ def write_connection_file(
     cfg["transport"] = transport
     cfg["signature_scheme"] = signature_scheme
     cfg["kernel_name"] = kernel_name
+    cfg.update(kwargs)
 
     # Only ever write this file as user read/writeable
     # This would otherwise introduce a vulnerability as a file has secrets
@@ -168,8 +177,8 @@ def write_connection_file(
 
 def find_connection_file(
     filename: str = "kernel-*.json",
-    path: Optional[Union[str, List[str]]] = None,
-    profile: Optional[str] = None,
+    path: str | list[str] | None = None,
+    profile: str | None = None,
 ) -> str:
     """find a connection file, and return its absolute path.
 
@@ -192,7 +201,9 @@ def find_connection_file(
     str : The absolute path of the connection file.
     """
     if profile is not None:
-        warnings.warn("Jupyter has no profiles. profile=%s has been ignored." % profile)
+        warnings.warn(
+            "Jupyter has no profiles. profile=%s has been ignored." % profile, stacklevel=2
+        )
     if path is None:
         path = [".", jupyter_runtime_dir()]
     if isinstance(path, str):
@@ -229,10 +240,10 @@ def find_connection_file(
 
 
 def tunnel_to_kernel(
-    connection_info: Union[str, KernelConnectionInfo],
+    connection_info: str | KernelConnectionInfo,
     sshserver: str,
-    sshkey: Optional[str] = None,
-) -> Tuple[Any, ...]:
+    sshkey: str | None = None,
+) -> tuple[Any, ...]:
     """tunnel connections to a kernel via ssh
 
     This will open five SSH tunnels from localhost on this machine to the
@@ -279,7 +290,7 @@ def tunnel_to_kernel(
     remote_ip = cf["ip"]
 
     if tunnel.try_passwordless_ssh(sshserver, sshkey):
-        password: Union[bool, str] = False
+        password: bool | str = False
     else:
         password = getpass("SSH Password for %s: " % sshserver)
 
@@ -307,9 +318,9 @@ port_names = ["%s_port" % channel for channel in ("shell", "stdin", "iopub", "hb
 class ConnectionFileMixin(LoggingConfigurable):
     """Mixin for configurable classes that work with connection files"""
 
-    data_dir: Union[str, Unicode] = Unicode()
+    data_dir: str | Unicode = Unicode()
 
-    def _data_dir_default(self):
+    def _data_dir_default(self) -> str:
         return jupyter_data_dir()
 
     # The addresses for the communication channels
@@ -326,7 +337,7 @@ class ConnectionFileMixin(LoggingConfigurable):
     _connection_file_written = Bool(False)
 
     transport = CaselessStrEnum(["tcp", "ipc"], default_value="tcp", config=True)
-    kernel_name: Union[str, Unicode] = Unicode()
+    kernel_name: str | Unicode = Unicode()
 
     context = Instance(zmq.Context)
 
@@ -338,7 +349,7 @@ class ConnectionFileMixin(LoggingConfigurable):
         to the Kernel, so be careful!""",
     )
 
-    def _ip_default(self):
+    def _ip_default(self) -> str:
         if self.transport == "ipc":
             if self.connection_file:
                 return os.path.splitext(self.connection_file)[0] + "-ipc"
@@ -348,7 +359,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             return localhost()
 
     @observe("ip")
-    def _ip_changed(self, change):
+    def _ip_changed(self, change: Any) -> None:
         if change["new"] == "*":
             self.ip = "0.0.0.0"  # noqa
 
@@ -361,16 +372,16 @@ class ConnectionFileMixin(LoggingConfigurable):
     control_port = Integer(0, config=True, help="set the control (ROUTER) port [default: random]")
 
     # names of the ports with random assignment
-    _random_port_names: Optional[List[str]] = None
+    _random_port_names: list[str] | None = None
 
     @property
-    def ports(self) -> List[int]:
+    def ports(self) -> list[int]:
         return [getattr(self, name) for name in port_names]
 
     # The Session to use for communication with the kernel.
     session = Instance("jupyter_client.session.Session")
 
-    def _session_default(self):
+    def _session_default(self) -> Session:
         from .session import Session
 
         return Session(parent=self)
@@ -420,10 +431,10 @@ class ConnectionFileMixin(LoggingConfigurable):
     # factory for blocking clients
     blocking_class = Type(klass=object, default_value="jupyter_client.BlockingKernelClient")
 
-    def blocking_client(self):
+    def blocking_client(self) -> BlockingKernelClient:
         """Make a blocking client connected to my kernel"""
         info = self.get_connection_info()
-        bc = self.blocking_class(parent=self)
+        bc = self.blocking_class(parent=self)  # type:ignore[operator]
         bc.load_connection_info(info)
         return bc
 
@@ -482,7 +493,7 @@ class ConnectionFileMixin(LoggingConfigurable):
 
         self.cleanup_connection_file()
 
-    def write_connection_file(self) -> None:
+    def write_connection_file(self, **kwargs: Any) -> None:
         """Write connection info to JSON dict in self.connection_file."""
         if self._connection_file_written and os.path.exists(self.connection_file):
             return
@@ -499,6 +510,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             control_port=self.control_port,
             signature_scheme=self.session.signature_scheme,
             kernel_name=self.kernel_name,
+            **kwargs,
         )
         # write_connection_file also sets default ports:
         self._record_random_port_names()
@@ -507,7 +519,7 @@ class ConnectionFileMixin(LoggingConfigurable):
 
         self._connection_file_written = True
 
-    def load_connection_file(self, connection_file: Optional[str] = None) -> None:
+    def load_connection_file(self, connection_file: str | None = None) -> None:
         """Load connection info from JSON dict in self.connection_file.
 
         Parameters
@@ -536,7 +548,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             See the connection_file spec for details.
         """
         self.transport = info.get("transport", self.transport)
-        self.ip = info.get("ip", self._ip_default())
+        self.ip = info.get("ip", self._ip_default())  # type:ignore[assignment]
 
         self._record_random_port_names()
         for name in port_names:
@@ -634,7 +646,7 @@ class ConnectionFileMixin(LoggingConfigurable):
             return f"{transport}://{ip}-{port}"
 
     def _create_connected_socket(
-        self, channel: str, identity: Optional[bytes] = None
+        self, channel: str, identity: bytes | None = None
     ) -> zmq.sugar.socket.Socket:
         """Create a zmq Socket and connect it to the kernel."""
         url = self._make_url(channel)
@@ -648,25 +660,25 @@ class ConnectionFileMixin(LoggingConfigurable):
         sock.connect(url)
         return sock
 
-    def connect_iopub(self, identity: Optional[bytes] = None) -> zmq.sugar.socket.Socket:
+    def connect_iopub(self, identity: bytes | None = None) -> zmq.sugar.socket.Socket:
         """return zmq Socket connected to the IOPub channel"""
         sock = self._create_connected_socket("iopub", identity=identity)
         sock.setsockopt(zmq.SUBSCRIBE, b"")
         return sock
 
-    def connect_shell(self, identity: Optional[bytes] = None) -> zmq.sugar.socket.Socket:
+    def connect_shell(self, identity: bytes | None = None) -> zmq.sugar.socket.Socket:
         """return zmq Socket connected to the Shell channel"""
         return self._create_connected_socket("shell", identity=identity)
 
-    def connect_stdin(self, identity: Optional[bytes] = None) -> zmq.sugar.socket.Socket:
+    def connect_stdin(self, identity: bytes | None = None) -> zmq.sugar.socket.Socket:
         """return zmq Socket connected to the StdIn channel"""
         return self._create_connected_socket("stdin", identity=identity)
 
-    def connect_hb(self, identity: Optional[bytes] = None) -> zmq.sugar.socket.Socket:
+    def connect_hb(self, identity: bytes | None = None) -> zmq.sugar.socket.Socket:
         """return zmq Socket connected to the Heartbeat channel"""
         return self._create_connected_socket("hb", identity=identity)
 
-    def connect_control(self, identity: Optional[bytes] = None) -> zmq.sugar.socket.Socket:
+    def connect_control(self, identity: bytes | None = None) -> zmq.sugar.socket.Socket:
         """return zmq Socket connected to the Control channel"""
         return self._create_connected_socket("control", identity=identity)
 
@@ -684,7 +696,7 @@ class LocalPortCache(SingletonConfigurable):
 
     def __init__(self, **kwargs: Any) -> None:
         super().__init__(**kwargs)
-        self.currently_used_ports: Set[int] = set()
+        self.currently_used_ports: set[int] = set()
 
     def find_available_port(self, ip: str) -> int:
         while True:
